/*

    Program: glgrab - test using output through OpenGL
    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer  -  Deskplayer.Com Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  | 
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/


#include <stdio.h>
#include <stdlib.h>

#include "GL/glut.h"

#include "bgrab.h"

/* Texture size */

#define WIDTH	256
#define HEIGHT	256

/* Grabbing size */
#define GWIDTH  256
#define GHEIGHT 256
#define DEPTH   3

/* Initial window size - resizable */
#define XWIDTH	256
#define XHEIGHT	256

/* Framegrabber structure */
static struct fgdevice fg;

/* For TV-set functions */
static int region=REGION_NTSC_CABLE;
static int channel=1;

/* Bitmap reversal variables */
static unsigned char *reverse, *curreverse, *curpos;

/* Idle function - just redraw */
static void idle(void)
{
 /* Print fps */
 if (fg.framecount==0) {
  printf ("FPS=%7.2f\n",fg_get_fps(&fg));
 }
 glutPostRedisplay();
}

/* Draw live video as textured polygon */    
static void draw( void )
{
 int i;
 
 glClear( GL_COLOR_BUFFER_BIT );
                              
 /* Update framebuffer with new image */
 if (fg_get_next_image(&fg)!=NULL) {

  /* Reverse pixel order */
  curpos=fg.current_image;
  curreverse=reverse;
  curreverse += (WIDTH*HEIGHT*DEPTH-1);
  for (i=0; i<(GWIDTH*GHEIGHT*DEPTH); i++) {
   *curreverse=*curpos;
   curpos++;
   curreverse--;
  }
    
  /* Setup texture */
  glBindTexture (GL_TEXTURE_2D, 1);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, WIDTH, HEIGHT, 0,
                GL_RGB, GL_UNSIGNED_BYTE, reverse );

  /* Draw polygon */
  glDepthFunc(GL_EQUAL);
  glEnable(GL_TEXTURE_2D);
  glBindTexture( GL_TEXTURE_2D, 1 );
  glPushMatrix();
   glBegin( GL_POLYGON );
    glTexCoord2f( 1.0, 0.0 );   glVertex2f( -1.0, -1.0 );
    glTexCoord2f( 0.0, 0.0 );   glVertex2f(  1.0, -1.0 );
    glTexCoord2f( 0.0, 1.0 );   glVertex2f(  1.0,  1.0 );
    glTexCoord2f( 1.0, 1.0 );   glVertex2f( -1.0,  1.0 );
   glEnd();
  glPopMatrix();
  glDisable(GL_TEXTURE_2D);
 } 

 glutSwapBuffers();
}
                                      
/* Keyboard controls */                                                                                                                                    
static void key( unsigned char key, int x, int y )
{
 (void) x;
 (void) y;
 switch (key) {
  case 27:
     if (fg_stop_grab_image(&fg)!=0) exit(-1); 	
     if (fg_close_device (&fg)!=0) exit(-1); 	
     exit(0);
     break;
    case 'f':
     channel++;
     if (fg_set_frequency (&fg,region,channel)!=0) exit(-1);
     break;
    case 'd':
     channel--;
     if (fg_set_frequency (&fg,region,channel)!=0) exit(-1);
     break;
 }
}

/* Windows resizing */
static void reshape( int width, int height )
{
 glViewport(0, 0, (GLint)width, (GLint)height);
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 glFrustum( -1.0, 1.0, -1.0, 1.0, 6.0, 20.0 );
 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity();
 glTranslatef( 0.0, 0.0, -6.0 );
}

/* Window visible or not */
void visible(int vis)
{
 if (vis == GLUT_VISIBLE)
  glutIdleFunc(idle);
 else
  glutIdleFunc(NULL);
}
            
            
/* ---- glgrab main */
            
int main (int argc, char *argv[])
{

 fprintf (stderr,"glgrab - libbgrab testprogram\n\n");

 glutInitWindowPosition(0, 0);
 glutInitWindowSize(XWIDTH, XHEIGHT);
 glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE );
        
 if (glutCreateWindow("glgrab") <= 0) {
  exit(-1);
 }
    
 if (argc>1) {
  if (fg_open_device (&fg, argv[1])!=0) exit(-1); 	
 } else {
  if (fg_open_device (&fg, "/dev/video")!=0) exit(-1); 	
 }
 
 if (fg_print_info (&fg)!=0) exit(-1);

 /* --- This will likely have to changed for YOUR setup ---------- */
 /* .............................................................. */

 if (fg_set_channel (&fg, CHANNEL_TUNER, VIDEOMODE_NTSC)!=0) exit(-1);
 if (fg_set_frequency (&fg, region, channel)!=0) exit(-1);

 /* .............................................................. */

 if (fg_start_grab_image(&fg, GWIDTH, GHEIGHT, FORMAT_RGB24)!=0) exit(-1);
 fg_set_fps_interval(&fg,100);

 /* Create reversion buffer */
 reverse=malloc(WIDTH*HEIGHT*DEPTH);
 
 glutReshapeFunc( reshape );
 glutKeyboardFunc( key );
 glutDisplayFunc( draw );
 glutVisibilityFunc(visible);
 
 glutMainLoop();

 exit(0);
}
