/*

    Program: rgbmix - mixes three seperate b/w input into a RGB image
                      Requires 3 Framegrabber cards!
    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer    -    Deskplayer Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |  
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/

/*
 * NOTE: Requires XWindows running at 16bit depth 565 weight.
 */

/* The image hold function requires to set the hold flag and the update flag */


#include <stdio.h>
#include <stdlib.h>

#include "bgrab.h"
#include "xutil.h"

#define WIDTH	640
#define HEIGHT	480
#define DEPTH 	1

#define XWIDTH	640
#define XHEIGHT	480
#define XDEPTH 	2

/* Brightness/Contrast setup ... values between 0 and 64000 */

static unsigned short brightness0	= 32000;
static unsigned short contrast0		= 32000;
static unsigned short brightness1	= 32000;
static unsigned short contrast1		= 32000;
static unsigned short brightness2	= 32000;
static unsigned short contrast2		= 32000;

int main (int argc, char *argv[])
{
 int i;
 struct fgdevice fg0;
 struct fgdevice fg1;
 struct fgdevice fg2;
 struct xwinbuffer xwin;
 int not_first_time=0;
 int in_loop=1;
 KeySym key;
 int brightness, contrast;
 int button;
 int mousex,mousey;
 int framecount=0;
 int hold0, hold1, hold2;
 int update0, update1, update2;
 unsigned char *oldframe0,*oldframe1,*oldframe2;
 unsigned char *curposframe0,*curposframe1,*curposframe2;
 unsigned short *curposscreen;

 fprintf (stderr,"rgbmix - libbgrab testprogram\n\n");
  
 if (setup_xwinbuffer (&xwin, argv[0], XWIDTH, XHEIGHT, XDEPTH)!=0) exit(-1);

 fprintf (stderr,"Opening /dev/video0\n");   
 if (fg_open_device (&fg0, "/dev/video0")!=0) exit(-1); 	
 if (fg_print_info (&fg0)!=0) exit(-1);

 fprintf (stderr,"Opening /dev/video1\n");   
 if (fg_open_device (&fg1, "/dev/video1")!=0) exit(-1); 	
 if (fg_print_info (&fg1)!=0) exit(-1);

 fprintf (stderr,"Opening /dev/video2\n");   
 if (fg_open_device (&fg2, "/dev/video2")!=0) exit(-1); 	
 if (fg_print_info (&fg2)!=0) exit(-1);

 /* These lines might require change depending on YOUR input */
 /* ........................................................ */
 if (fg_set_channel (&fg0,CHANNEL_COMPOSITE,VIDEOMODE_NTSC)!=0) exit(-1);
 if (fg_set_channel (&fg1,CHANNEL_COMPOSITE,VIDEOMODE_NTSC)!=0) exit(-1);
 if (fg_set_channel (&fg2,CHANNEL_COMPOSITE,VIDEOMODE_NTSC)!=0) exit(-1);
 
 if (fg_start_grab_image(&fg0, WIDTH, HEIGHT, FORMAT_GREY)!=0) exit(-1);
 if (fg_start_grab_image(&fg1, WIDTH, HEIGHT, FORMAT_GREY)!=0) exit(-1);
 if (fg_start_grab_image(&fg2, WIDTH, HEIGHT, FORMAT_GREY)!=0) exit(-1);

 fg_set_fps_interval(&fg0,100);

 /* Setup grabbers input */
 fg_set_setting(&fg0,SETTING_BRIGHTNESS,brightness0);
 fg_set_setting(&fg0,SETTING_CONTRAST,contrast0);
 fg_set_setting(&fg1,SETTING_BRIGHTNESS,brightness1);
 fg_set_setting(&fg1,SETTING_CONTRAST,contrast1);
 fg_set_setting(&fg2,SETTING_BRIGHTNESS,brightness2);
 fg_set_setting(&fg2,SETTING_CONTRAST,contrast2);

 /* Get memory for image storage */
 oldframe0=malloc(WIDTH*HEIGHT*DEPTH);
 oldframe1=malloc(WIDTH*HEIGHT*DEPTH);
 oldframe2=malloc(WIDTH*HEIGHT*DEPTH);

 /* Initialize store variables */
 hold0=0;
 hold1=0;
 hold2=0;
 update0=0;
 update1=0;
 update2=0;
 
 /* View loop */
 while (in_loop) { 

  /* Update framebuffer 0 with new image */
  if (fg_get_next_image(&fg0)!=NULL) {
   /* Update framebuffer 1 with new image */
   if (fg_get_next_image(&fg1)!=NULL) {
    /* Update framebuffer 2 with new image */
    if (fg_get_next_image(&fg2)!=NULL) {

     /* Count images */
     framecount++;

     /* Store images */
     if (update0) {
      memcpy(oldframe0,fg0.current_image,fg0.image_size);
      update0=0;
     }
     if (update1) {
      memcpy(oldframe1,fg1.current_image,fg1.image_size);
      update1=0;
     }
     if (update2) {
      memcpy(oldframe2,fg2.current_image,fg2.image_size);
      update2=0;
     }
  
     /* Mix live or stored images to a single RGB565 screenbuffer */
     /* Set pointers to live or store images */
     if (hold0) { 
      curposframe0=oldframe0;
     } else {
      curposframe0=fg0.current_image;
     }
     if (hold1) { 
      curposframe1=oldframe1;
     } else {
      curposframe1=fg1.current_image;
     }
     if (hold2) { 
      curposframe2=oldframe2;
     } else {
      curposframe2=fg2.current_image;
     }
     curposscreen=(unsigned short *)xwin.shmimage->data;
     /* Copy image to screenbuffer mixing rgb channels */
     for (i=0; i<fg0.image_size; i++) {
      *curposscreen =  ((unsigned short)*curposframe0 >> 3) | 
                      (((unsigned short)*curposframe1 & (0xff-0x03)) << 3) | 
                      (((unsigned short)*curposframe2 & (0xff-0x07)) << 8) ;
      curposscreen++;
      curposframe0++;
      curposframe1++;
      curposframe2++;
     }

     /* Draw it */
     sync_xwinbuffer(&xwin);

    }
   }
  }

  /* Poll keyboard */
  if (keypoll_xwinbuffer (&xwin, &key, &button, &mousex, &mousey)) {
   switch (key) {
    case XK_q:
     in_loop=0;
     break; 
    case XK_1:
     hold0 ^= 1;
     if (hold0==1) update0=1;
     break; 
    case XK_2:
     hold1 ^= 1;
     if (hold1==1) update1=1;
     break; 
    case XK_3:
     hold2 ^= 1;
     if (hold2==1) update2=1;
     break; 

    case XK_c:
     contrast=fg_get_setting(&fg0,SETTING_CONTRAST)-64;
     fprintf (stderr,"fg0: Contrast set to %i\n",contrast);
     fg_set_setting(&fg0,SETTING_CONTRAST,contrast);
     break; 
    case XK_v:
     contrast=fg_get_setting(&fg0,SETTING_CONTRAST)+64;
     fprintf (stderr,"fg0: Contrast set to %i\n",contrast);
     fg_set_setting(&fg0,SETTING_CONTRAST,contrast);
     break; 
    case XK_b:
     brightness=fg_get_setting(&fg0,SETTING_BRIGHTNESS)-64;
     fprintf (stderr,"fg0: Brightness set to %i\n",brightness);
     fg_set_setting(&fg0,SETTING_BRIGHTNESS,brightness);
     break; 
    case XK_n:
     brightness=fg_get_setting(&fg0,SETTING_BRIGHTNESS)+64;
     fprintf (stderr,"fg0: Brightness set to %i\n",brightness);
     fg_set_setting(&fg0,SETTING_BRIGHTNESS,brightness);
     break; 

    case XK_a:
     contrast=fg_get_setting(&fg1,SETTING_CONTRAST)-64;
     fprintf (stderr,"fg1: Contrast set to %i\n",contrast);
     fg_set_setting(&fg1,SETTING_CONTRAST,contrast);
     break; 
    case XK_s:
     contrast=fg_get_setting(&fg1,SETTING_CONTRAST)+64;
     fprintf (stderr,"fg1: Contrast set to %i\n",contrast);
     fg_set_setting(&fg1,SETTING_CONTRAST,contrast);
     break; 
    case XK_d:
     brightness=fg_get_setting(&fg1,SETTING_BRIGHTNESS)-64;
     fprintf (stderr,"fg1: Brightness set to %i\n",brightness);
     fg_set_setting(&fg1,SETTING_BRIGHTNESS,brightness);
     break; 
    case XK_f:
     brightness=fg_get_setting(&fg1,SETTING_BRIGHTNESS)+64;
     fprintf (stderr,"fg1: Brightness set to %i\n",brightness);
     fg_set_setting(&fg1,SETTING_BRIGHTNESS,brightness);
     break; 

    case XK_g:
     contrast=fg_get_setting(&fg2,SETTING_CONTRAST)-64;
     fprintf (stderr,"fg2: Contrast set to %i\n",contrast);
     fg_set_setting(&fg2,SETTING_CONTRAST,contrast);
     break; 
    case XK_h:
     contrast=fg_get_setting(&fg2,SETTING_CONTRAST)+64;
     fprintf (stderr,"fg2: Contrast set to %i\n",contrast);
     fg_set_setting(&fg2,SETTING_CONTRAST,contrast);
     break; 
    case XK_j:
     brightness=fg_get_setting(&fg2,SETTING_BRIGHTNESS)-64;
     fprintf (stderr,"fg2: Brightness set to %i\n",brightness);
     fg_set_setting(&fg2,SETTING_BRIGHTNESS,brightness);
     break; 
    case XK_k:
     brightness=fg_get_setting(&fg2,SETTING_BRIGHTNESS)+64;
     fprintf (stderr,"fg2: Brightness set to %i\n",brightness);
     fg_set_setting(&fg2,SETTING_BRIGHTNESS,brightness);
     break; 

   }
  }
  
  /* Print fps */
  if (fg0.framecount==0) {
   if (not_first_time) {
    printf ("FPS=%7.2f\n",fg_get_fps(&fg0));
   } else {
    not_first_time=1; 
   } 
  } 
 }           

 if (close_xwinbuffer(&xwin)!=0) exit(-1);
 
 if (fg_stop_grab_image(&fg0)!=0) exit(-1); 	
 if (fg_close_device (&fg0)!=0) exit(-1); 	
 if (fg_stop_grab_image(&fg1)!=0) exit(-1); 	
 if (fg_close_device (&fg1)!=0) exit(-1); 	
 if (fg_stop_grab_image(&fg2)!=0) exit(-1); 	
 if (fg_close_device (&fg2)!=0) exit(-1); 	

 free(oldframe0);
 free(oldframe1);
 free(oldframe2);
 
 exit(0);
}
