/*

    Program: testabsdiff - testprogram for MMX assembly routines.
    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer  -  Deskplayer.Com Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/

/*
 * NOTE: Requires XWindows running at 16bit depth 565 weight.
 */


#include <stdio.h>
#include <stdlib.h>

#include "bgrab.h"
#include "xutil.h"
#include "filters.h"

#define WIDTH	640
#define HEIGHT	480
#define DEPTH 	1

#define XWIDTH	640
#define XHEIGHT	480
#define XDEPTH 	2

int main (int argc, char *argv[])
{
 int i;
 struct fgdevice fg;
 struct xwinbuffer xwin;
 int not_first_time=0;
 int in_loop=1;
 KeySym key;
 int button;
 int mousex,mousey;
 int contrast;
 int brightness;
 int framecount=0;
 int threshold_value=0;
 unsigned char *lastimage, *curposframe;
 unsigned char *outputimage, *outputimage2;
 unsigned short *curposscreen;

 fprintf (stderr,"testabsdiff - libbgrab testprogram\n\n");
  
 if (setup_xwinbuffer (&xwin, argv[0], XWIDTH, XHEIGHT, XDEPTH)!=0) exit(-1);
   
 if (argc>1) {
  if (fg_open_device (&fg, argv[1])!=0) exit(-1); 	
 } else {
  if (fg_open_device (&fg, "/dev/video")!=0) exit(-1); 	
 }
 
 if (fg_print_info (&fg)!=0) exit(-1);

 /* This line might require change depending on YOUR input */
 /* ...................................................... */
 if (fg_set_channel (&fg,CHANNEL_TUNER,VIDEOMODE_NTSC)!=0) exit(-1);
 fg_set_frequency(&fg,1,1);
  
 if (fg_start_grab_image(&fg, WIDTH, HEIGHT, FORMAT_GREY)!=0) exit(-1);

 fg_set_fps_interval(&fg,100);

 /* Allocate memory for image storage */
 lastimage=malloc(WIDTH*HEIGHT*DEPTH);
 outputimage=malloc(WIDTH*HEIGHT*DEPTH);
 outputimage2=malloc(WIDTH*HEIGHT*DEPTH);

 /* View loop */
 while (in_loop) { 

  /* Update framebuffer with new image */
  if (fg_get_next_image(&fg)!=NULL) {

   /* Count images */
   framecount++;

   /* Use MMX optimized assembly routine to calculate absolute difference */
   absdiff(fg.current_image,lastimage,outputimage,fg.image_size);
 
   
   if (threshold_value>0) {
    /* Use MMX optimized assembly routine to calculate threshold */
    threshold(outputimage, outputimage2, fg.image_size, threshold_value);
   }
      
   /* Store current image in buffer for next iteration */
   memcpy((char *)lastimage,(char *)fg.current_image,fg.image_size);

   /* Copy output image to screenbuffer */
   if (threshold_value>0) {
    curposframe=outputimage2;
   } else {
    curposframe=outputimage;
   }
   curposscreen=(unsigned short *)xwin.shmimage->data;
   for (i=0; i<fg.image_size; i++) {
    *curposscreen = fg.y8_to_rgb565[(int)*curposframe];
    curposscreen++;
    curposframe++;
   }      

   if (framecount<200) {
    /* Say something ... */
    drawstring (&xwin, 8, 8, "Press Q to quit", framecount);
   }
        
   /* Draw it */
   sync_xwinbuffer(&xwin);
  }

  /* Poll keyboard */
  if (keypoll_xwinbuffer (&xwin, &key, &button, &mousex, &mousey)) {
   switch (key) {
    case XK_q:
     in_loop=0;
     break; 

    case XK_t:
     if (threshold_value>0) {
      threshold_value=0;
     } else {
      threshold_value=64;
     }
     break;
     
    case XK_c:
     contrast=fg_get_setting(&fg,SETTING_CONTRAST)-64;
     fprintf (stderr,"Contrast set to %i\n",contrast);
     fg_set_setting(&fg,SETTING_CONTRAST,contrast);
     break; 

    case XK_v:
     contrast=fg_get_setting(&fg,SETTING_CONTRAST)+64;
     fprintf (stderr,"Contrast set to %i\n",contrast);
     fg_set_setting(&fg,SETTING_CONTRAST,contrast);
     break; 

    case XK_b:
     brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)-64;
     fprintf (stderr,"Brightness set to %i\n",brightness);
     fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
     break; 

    case XK_n:
     brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)+64;
     fprintf (stderr,"Brightness set to %i\n",brightness);
     fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
     break; 
   }
  }
  
  /* Print fps */
  if (fg.framecount==0) {
   if (not_first_time) {
    printf ("FPS=%7.2f\n",fg_get_fps(&fg));
   } else {
    not_first_time=1; 
   } 
  } 
 }           

 if (close_xwinbuffer(&xwin)!=0) exit(-1);
 
 if (fg_stop_grab_image(&fg)!=0) exit(-1); 	

 if (fg_close_device (&fg)!=0) exit(-1); 	
 
 exit(0);
}
