/*

    Program: expograb - delaygrab version for Expo2000
    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer    -    Deskplayer Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/

/*
 * NOTE: Requires XWindows running at 16bit depth 565 weight.
 *       Requires a lot of memory.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#include "bgrab.h"
#include "xutil.h"
#include "netsocket.h"

#define WIDTH	640
#define HEIGHT	480
#define DEPTH 	2

#define BLOCKSIZE	4
#define DELAYMAPWIDTH   (WIDTH/BLOCKSIZE)
#define DELAYMAPHEIGHT	(HEIGHT/BLOCKSIZE)
#define DELAYMAPSIZE	(DELAYMAPHEIGHT*DELAYMAPWIDTH)

#define MAXDELAYMAPNUM	64*1024

#define QUEUEDEPTH	128

#define BUTTON_MAX 	48

/* Framegrabber setup */

static int brightness;
static int contrast;
static int color;

/* Globals */

int x,y,i,xyoff,v;
struct fgdevice fg;
struct xwinbuffer xwin;
int not_first_time=0;
int in_loop=1;
KeySym key;
int button;
int mousex,mousey;
char *imagequeue,*curqueue;
int curqueuenum;
int *delaymap=NULL,*curdelaymap;
char *curpos,*curimage;
int curposnum;
double d;

/* Network-adjustable numbers */

int target_reference;
int button_number[BUTTON_MAX];
int user_number=0;
int network_number=0;

#define LINE_LEN	256

static char newfilename[LINE_LEN]="";
static char oldfilename[LINE_LEN]="";

static int INPUTCHANNEL,
           PORT, VIDEOMODE,
	   MODECHANGE,
	   REGION,
	   INDEX, OVERLAY, OVERLAYBG, OVERLAYFG, SHOWMAP;
static char VIDEODEV[LINE_LEN];

void set_defaults()
{
 INPUTCHANNEL=CHANNEL_COMPOSITE;
 PORT=1334;
 strcpy (VIDEODEV,"/dev/video");
 VIDEOMODE=VIDEOMODE_PAL;
 MODECHANGE=0;
 REGION=1;
 INDEX=1;
 OVERLAY=0;
 OVERLAYBG=0;
 OVERLAYFG=255*256*256+255*256+255;
 SHOWMAP=0;
}

/* Configuration file settings */

#define DEFAULT_FILENAME	"expograb.cnf"

void SaveDefaultSettings()
{
 FILE *deffile;
 
 deffile=fopen(DEFAULT_FILENAME,"w");

 fprintf(deffile,"# Framegrabber controls\n");
 fprintf(deffile,"contrast	%i\n",contrast);
 fprintf(deffile,"brightness	%i\n",brightness);
 fprintf(deffile,"color		%i\n",color);

 fclose(deffile);

 fprintf(stderr,"Settings saved ...\n");
}

void LoadDefaultSettings()
{
 FILE *deffile;
 char aline[LINE_LEN];
 char tag[LINE_LEN];
 int value;
  
 /* Open file */
 if ((deffile=fopen(DEFAULT_FILENAME,"r"))!=NULL) {
 
  /* Scan file for tags/value pairs */
  /* Fetch a line */
  fgets(aline,LINE_LEN,deffile);  
  while (!feof(deffile)) {
   /* Test line we read */   
   if (aline[0]!='#') {
    /* Scan tag-names and their value */
    sscanf(aline,"%s %i",tag,&value);
   
    printf ("Tag=%s Value=%i\n",tag,value);
    
    if (strncasecmp(aline,"contrast",8)==0) {
      contrast=value;
    }
    if (strncasecmp(aline,"brightness",8)==0) {
      brightness=value;
    }
    if (strncasecmp(aline,"color",5)==0) {
      color=value;
    }
   }  
   /* Get next line */
   fgets(aline,LINE_LEN,deffile);  
  }
 
  /* Done - now close */
  fclose(deffile);

 } else {
  fprintf (stderr,"Cannot load %s.\n",DEFAULT_FILENAME);
 } 
}


int LoadDelaymap (char *pathname)
{
 int i,xsize,ysize,numcolors;
 char filename[LINE_LEN];
 char header[LINE_LEN];
 char errormessage[LINE_LEN];
 char aline[LINE_LEN];
 FILE *infile;
 unsigned char tempbuffer[DELAYMAPSIZE],*curtempbuffer;
  
 /* Message */
 fprintf (stderr,"Loading delaymap from %s:\n",pathname);
 /* Load font textures */
 /* Build filename */
 sprintf(filename,"maps/%s.pgm",pathname);
 /* Open and read file */
 if ((infile=fopen(filename,"r"))==NULL) {
  sprintf(errormessage,"fopen(%s)",filename);
  perror(errormessage);
  return(-1);
 }
 /* Read header */
 fgets(aline,LINE_LEN,infile);
 sscanf(aline,"%s",header);
 /* Dump comment lines */
 fgets(aline,LINE_LEN,infile);
 while (aline[0]=='#') {
  fprintf (stderr,"Comment: %s",aline);
  fgets(aline,LINE_LEN,infile);
 } 
 /* Convert size from line */
 sscanf(aline,"%i %i",&xsize,&ysize);
 /* Convert depth from line */ 
 fgets(aline,LINE_LEN,infile);
 sscanf(aline,"%i",&numcolors);

 /* Progress */
 fprintf (stderr,"%s: %s  %ix%i  %i colors\n",filename,header,xsize,ysize,numcolors);

 /* Check image size */
 if ((xsize==DELAYMAPWIDTH) && (ysize==DELAYMAPHEIGHT)) {
  /* Read data */
  fread(tempbuffer,DELAYMAPSIZE,1,infile);
  /* Adjust range */
  curtempbuffer=tempbuffer;
  curdelaymap=delaymap;
  for (i=0; i<(DELAYMAPSIZE); i++) {
   *curdelaymap = (int)(*curtempbuffer >> 1);
   curtempbuffer++;
   curdelaymap++;
  }
 } else {
  fprintf (stderr,"Error reading %s: Bad Size!\n",filename);
 } // size check 

 fclose(infile);
 
 return(0); 
}

/* ------------- */

int main (int argc, char *argv[])
{
 int i,j,x1,y1,x2,y2;
 int framecount=0;
 struct netthread nt;
 int reload;
 char aline[LINE_LEN];

 fprintf (stderr,"expograb - Expo2000 exhibition version of delaygrab\n\n");

 /* Initialize variables */
 set_defaults();
 
 /* Parse command line */
 for (i=1; i<argc; i++) {
  char	*arg;
  if (argv[i][0]!='-') break;
  arg = argv[i];
  arg++;
  if (*arg == '-') arg++;
  if (strncmp(arg,"port",4)==0) {
   i++; 
   if (i<=argc) {
    PORT = atoi(argv[i]);
   }
  } else if (strncmp(arg,"channel",7)==0) {
   i++; 
   if (i<=argc) {
    INPUTCHANNEL = atoi(argv[i]);
   }
  } else if (strncmp(arg,"region",6)==0) {
   i++; 
   if (i<=argc) {
    REGION = atoi(argv[i]);
   }
  } else if (strncmp(arg,"index",5)==0) {
   i++; 
   if (i<=argc) {
    INDEX = atoi(argv[i]);
   }
  } else if (strncmp(arg,"videodev",8)==0) {
   i++; 
   if (i<=argc) {
    strcpy (VIDEODEV,argv[i]);
   }
  } else if (strncmp(arg,"ntsc",4)==0) {
   VIDEOMODE=VIDEOMODE_NTSC;
  } else if (strncmp(arg,"secam",5)==0) {
   VIDEOMODE=VIDEOMODE_SECAM;
  } else if (strncmp(arg,"modechange",10)==0) {
   MODECHANGE=1;
  } else if (strncmp(arg,"overlay",7)==0) {
   OVERLAY = 1;
  } else if (strncmp(arg,"showmap",7)==0) {
   SHOWMAP = 1;
  } else if (strncmp(arg,"bgcolor",7)==0) {
   i++; 
   if (i<=argc) {
    OVERLAYBG = atoi(argv[i]);
   }
  } else if (strncmp(arg,"fgcolor",7)==0) {
   i++; 
   if (i<=argc) {
    OVERLAYFG = atoi(argv[i]);
   }
  } else {
   /* Command line help */
   printf ("Command line parameters:\n");
   printf ("\nVideo settings:\n");
   printf (" -channel #         input channel (default: %i)\n",INPUTCHANNEL);
   printf (" -videodev txt      video device (default: %s)\n",VIDEODEV);
   printf (" -ntsc              use NTSC video (default: pal)\n");
   printf (" -secam             use SECAM video (default: pal)\n");
   printf (" -region #		set the frequency region (default: %i)\n",REGION);
   printf (" -index #		set the frequency index (default: %i)\n",INDEX);
   printf ("\nNetwork settings:\n");
   printf (" -port #            listening port (default: %i)\n",PORT);
   printf ("\nOutput settings:\n");
   printf (" -modechange        change videomode to %ix%i (default: off)\n",WIDTH,HEIGHT);
   printf (" -overlay           turn on graphics-overlay  (default: off)\n");
   printf (" -showmap           turn on map number display (default: off)\n");
   printf (" -bgcolor #         number is background color (default: grey)\n");
   printf (" -fgcolor #         number is foreground color (default: red)\n");
   printf ("                    [bg/fg are optional 24bit RGB unsigned integer numbers]\n");
   printf ("                    [white=16777215 blue=255 green=65280 red=16711680]\n");
   exit(0);
  }
 }

 /* Create delay map array */
 delaymap=(int *)malloc(DELAYMAPSIZE*sizeof(int));

 /* Load initial Delaymap */
 LoadDelaymap("map1");
 target_reference=1;
 
 /* Open screen */
 if (setup_xwinbuffer (&xwin, argv[0], WIDTH, HEIGHT, DEPTH)!=0) exit(-1);

 /* Switch resolution */
 if (MODECHANGE) {
  videoModeChange (&xwin,WIDTH,HEIGHT);
 }
   
 /* Open framegrabber */ 
 if (fg_open_device (&fg, VIDEODEV)!=0) exit(-1); 	
 
 /* Setup framegrabber */
 if (fg_print_info (&fg)!=0) exit(-1);
 if (fg_set_channel (&fg,INPUTCHANNEL,VIDEOMODE)!=0) exit(-1);
 fg_set_fps_interval(&fg,100);
 if (fg_set_frequency (&fg,REGION,INDEX)!=0) exit(-1);
 
 /* Read settings from card */
 contrast=fg_get_setting(&fg,SETTING_CONTRAST)-256;
 brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)-256;
 color=fg_get_setting(&fg,SETTING_COLOUR)-256;

 /* Load defaults */
 LoadDefaultSettings();
 
 /* Set settings */
 fg_set_setting(&fg,SETTING_CONTRAST,contrast);
 fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
 fg_set_setting(&fg,SETTING_COLOUR,color);

 /* Start grabbing */
 if (fg_start_grab_image(&fg, WIDTH, HEIGHT, FORMAT_RGB565)!=0) exit(-1);

 /* Create storage queue */ 
 if ((imagequeue=(char *)malloc(QUEUEDEPTH*fg.image_size))==NULL) {
  fprintf(stderr,"Not enough memory for a %i frame queue (%i bytes) - exiting.\n",QUEUEDEPTH,QUEUEDEPTH*fg.image_size);
  in_loop=0;
 } else {
  curqueue=imagequeue;
  curqueuenum=0;
 }

 /* Setup network */
 if (start_network_socket(&nt,PORT,BE_VERBOSE)!=0) exit(-1);

 /* View loop */
 while (in_loop) { 

  /* Update framebuffer with new image */
  if (fg_get_next_image(&fg)!=NULL) {
   /* Count the frames */
   framecount++;

   /* Update queue pointer */
   if (curqueuenum==0) {
    curqueuenum=QUEUEDEPTH-1;
    curqueue = imagequeue;
    curqueue += (fg.image_size*(QUEUEDEPTH-1));
   } else {
    curqueuenum--;
    curqueue -= (fg.image_size);
   }
   
   /* Copy image to queue */
   memcpy((char *)curqueue,(char *)fg.current_image,fg.image_size);

   /* Copy image blockwise to screenbuffer */
   curdelaymap=delaymap;
   for (y=0; y<DELAYMAPHEIGHT; y++) {
    for (x=0; x<DELAYMAPWIDTH; x++) {
     curposnum=((curqueuenum + (*curdelaymap)) % QUEUEDEPTH);
     xyoff= x*(BLOCKSIZE*DEPTH) + y*(BLOCKSIZE*WIDTH*DEPTH);
     // source
     curpos=imagequeue;
     curpos += (fg.image_size*curposnum);
     curpos += xyoff;
     // target
     curimage = (char *)xwin.shmimage->data;
     curimage += xyoff;
     // copy block
     for (i=0; i<BLOCKSIZE; i++) {
      memcpy(curimage,curpos,(BLOCKSIZE*2));
      curpos += (WIDTH*DEPTH);
      curimage += (WIDTH*DEPTH);
     }
     curdelaymap++;
    }
   }
   
   /* Draw map number */
   if (SHOWMAP) {
    sprintf (aline,"%i",target_reference);
    x1=(640-5*8)/2; x2=x1+5*8;
    y1=10; y2=10+8+4;
    drawbox (&xwin,x1,y1,x2,y2,OVERLAYBG);
    hline (&xwin,x1,x2,y1,OVERLAYFG);
    hline (&xwin,x1,x2,y2,OVERLAYFG);
    vline (&xwin,x1,y1,y2,OVERLAYFG);
    vline (&xwin,x2,y1,y2,OVERLAYFG);
    x1=(640-strlen(aline)*8)/2;
    y1=13;
    drawstring (&xwin,x1,y1,aline,OVERLAYFG);
   }
    
   /* Draw overlay */
   if (OVERLAY) {
    /* Draw left bar */
    x1=10; x2=10+16;
    y1=10; y2=430;
    drawbox (&xwin,x1,y1,x2,y2,OVERLAYBG);
    hline (&xwin,x1,x2,y1,OVERLAYFG);
    hline (&xwin,x1,x2,y2,OVERLAYFG);
    vline (&xwin,x1,y1,y2,OVERLAYFG);
    vline (&xwin,x2,y1,y2,OVERLAYFG);
    /* Draw left bar inside */
    x1=10+2; x2=10+16-2;
    y1=12+(int)((float)(430-4-10)*0.01*(float)(100-network_number)); y2=430-2;
    drawbox (&xwin,x1,y1,x2,y2,OVERLAYFG);
    /* Draw left bar marking */
    x1=10+16+4; 
    y1=10-4;
    drawstring (&xwin,x1-1,y1-1,"100",OVERLAYBG);
    drawstring (&xwin,x1,y1,"100",OVERLAYFG);
    y1=10+420/2-4;
    drawstring (&xwin,x1-1,y1-1,"50",OVERLAYBG);
    drawstring (&xwin,x1,y1,"50",OVERLAYFG);
    /* Draw right bar */
    x1=WIDTH-10-16; x2=WIDTH-10;
    y1=10; y2=430;
    drawbox (&xwin,x1,y1,x2,y2,OVERLAYBG);
    hline (&xwin,x1,x2,y1,OVERLAYFG);
    hline (&xwin,x1,x2,y2,OVERLAYFG);
    vline (&xwin,x1,y1,y2,OVERLAYFG);
    vline (&xwin,x2,y1,y2,OVERLAYFG);
    /* Draw right bar inside */
    x1=WIDTH-10-16+2; x2=WIDTH-10-2;
    y1=12+(int)((float)(430-4-10)*0.01*(float)(100-user_number)); y2=430-2;
    drawbox (&xwin,x1,y1,x2,y2,OVERLAYFG);
    /* Draw right bar marking */
    x1=WIDTH-10-16-4-8*3; 
    y1=10-4;
    drawstring (&xwin,x1-1,y1-1,"100",OVERLAYBG);
    drawstring (&xwin,x1,y1,"100",OVERLAYFG);
    x1=WIDTH-10-16-4-8*2; 
    y1=10+420/2-4;
    drawstring (&xwin,x1-1,y1-1,"50",OVERLAYBG);
    drawstring (&xwin,x1,y1,"50",OVERLAYFG);
    /* Draw ticks */
    for (j=10; j<100; j += 10) {
     y1=12+(int)((float)(430-4-10)*0.01*(float)(j)); y2=430-2;
     drawpixel (&xwin,10+1,y1,OVERLAYFG);
     drawpixel (&xwin,10+16-1,y1,OVERLAYFG);
     drawpixel (&xwin,WIDTH-10-1,y1,OVERLAYFG);
     drawpixel (&xwin,WIDTH-10-16+1,y1,OVERLAYFG);
    }
        
    /* Draw button bar */
    x1=10; y1=430; x2=WIDTH-10; y2=470;
    drawbox (&xwin,x1,y1,x2,y2,OVERLAYBG);
    hline (&xwin,x1,x2,y1,OVERLAYFG);
    hline (&xwin,x1,x2,y2,OVERLAYFG);
    vline (&xwin,x1,y1,y2,OVERLAYFG);
    vline (&xwin,x2,y1,y2,OVERLAYFG);
    /* Draw text in button bar */
    x1=10; y1=430; x2=x1+7*8+2; y2=y1+8+2;
    drawbox (&xwin,x1,y1,x2,y2,OVERLAYFG);
    drawstring (&xwin,x1+1,y1+2,"NETWORK",OVERLAYBG);
    x1=WIDTH-12-4*8; y1=430; x2=x1+4*8+2; y2=y1+8+2;
    drawbox (&xwin,x1,y1,x2,y2,OVERLAYFG);
    drawstring (&xwin,x1+1,y1+2,"USER",OVERLAYBG);
    /* Draw buttons */
    for (i=0; i<BUTTON_MAX; i++) {
     /* Boxes for buttons */
     x1=WIDTH/2-12*20+20*(i % 24)+4; x2=x1+12;
     y1=HEIGHT-2*20-10+20*(i / 24)+4; y2=y1+12;
     if (button_number[i]) {
      drawbox (&xwin,x1+2,y1+2,x2-2,y2-2,OVERLAYFG);
     }   
     hline (&xwin,x1,x2,y1,OVERLAYFG);
     hline (&xwin,x1,x2,y2,OVERLAYFG);
     vline (&xwin,x1,y1,y2,OVERLAYFG);
     vline (&xwin,x2,y1,y2,OVERLAYFG);
    }
   }

   /* Draw it */
   sync_xwinbuffer(&xwin);
  }

  /* Check if we have a new frame */
  shared_data_lock(&nt);
  reload=(strcmp(newfilename,oldfilename)!=0);
  shared_data_unlock(&nt);
  /* If we do, load map */
  if (reload) {
   strcpy(oldfilename,newfilename);
   LoadDelaymap(newfilename);
  }

  /* Poll keyboard */
  if (keypoll_xwinbuffer (&xwin, &key, &button, &mousex, &mousey)) {
   switch (key) {

    case XK_q:
     in_loop=0;
     break; 

    case XK_c:
     contrast=fg_get_setting(&fg,SETTING_CONTRAST)-256;
     fprintf (stderr,"Contrast set to %i\n",contrast);
     fg_set_setting(&fg,SETTING_CONTRAST,contrast);
     break; 
   
    case XK_v:
     contrast=fg_get_setting(&fg,SETTING_CONTRAST)+256;
     fprintf (stderr,"Contrast set to %i\n",contrast);
     fg_set_setting(&fg,SETTING_CONTRAST,contrast);
     break; 
   
    case XK_b:
     brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)-256;
     fprintf (stderr,"Brightness set to %i\n",brightness);
     fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
     break; 
   
    case XK_n:
     brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)+256;
     fprintf (stderr,"Brightness set to %i\n",brightness);
     fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
     break; 

    case XK_i:
     color=fg_get_setting(&fg,SETTING_COLOUR)-256;
     fprintf (stderr,"Color set to %i\n",color);
     fg_set_setting(&fg,SETTING_COLOUR,color);
     break; 
   
    case XK_o:
     color=fg_get_setting(&fg,SETTING_COLOUR)+256;
     fprintf (stderr,"Color set to %i\n",color);
     fg_set_setting(&fg,SETTING_COLOUR,color);
     break; 
   
    case XK_s:
     SaveDefaultSettings();
     break; 

   }
  }
  
  /* Print fps */
  if (fg.framecount==0) {
   if (not_first_time) {
    printf ("FPS=%7.2f\n",fg_get_fps(&fg));
   } else {
    not_first_time=1; 
   } 
  } 
 }           

 /* Shutdown network */
 if (stop_network_socket(&nt)!=0) exit(-1);

 /* Reset video mode */
 if (MODECHANGE) {
  videoModeReset (&xwin);
 }

 /* Close window */
 if (close_xwinbuffer(&xwin)!=0) exit(-1);
 
 /* Close framegrabber */
 if (fg_stop_grab_image(&fg)!=0) exit(-1); 	
 if (fg_close_device (&fg)!=0) exit(-1); 	
 
 /* Cleanup memory */
 free (delaymap);
 free (imagequeue);

 exit(0);
}

/* Routine use by network thread to process commands */
int command_to_output(char *inbuffer, char* outbuffer, int outbufferlen)
{
 int i,imax;
 int target_selection;
 
 /* Mark for NO map selection */
 target_selection=0;
 /* Scan for different input variables by checking the first character */
 if (inbuffer!=NULL) {
  switch (inbuffer[0]) {
   case 'n':
   case 'N':
    /* Read number */
    sscanf (&inbuffer[1],"%i",&network_number);
    /* Just return input number*/
    snprintf (outbuffer,outbufferlen,"%i\n",network_number);
    break;
    
   case 'u':
   case 'U':
    /* Read number */
    sscanf (&inbuffer[1],"%i",&user_number);
    /* Just return input number*/
    snprintf (outbuffer,outbufferlen,"%i\n",user_number);
    break;

   case 'k':
   case 'K':
    imax = strlen(&inbuffer[1]);
    if (imax>BUTTON_MAX) imax=48;
    for (i=0; i<imax; i++) {
     button_number[i]=(inbuffer[i+1]=='1');
    }
    /* Just return input string */
    snprintf (outbuffer,outbufferlen,"%s\n",&inbuffer[1]);
    break;
   
   case 'M':
   case 'm':
    /* Read number */
    sscanf (&inbuffer[1],"%i",&target_selection);
    break;
    
   default:
    /* Read number */
    target_selection=0;
    sscanf (inbuffer,"%i",&target_selection);
    break;
  }

  /* Convert input command to output string */
  if ((target_selection>0) && (target_selection<=MAXDELAYMAPNUM)) {
    target_reference=target_selection;
    /* Just return number */
    snprintf (outbuffer,outbufferlen,"%i\n",target_selection);
    /* Set new filename */
    sprintf(newfilename,"map%i",target_selection);
   return(0);
  }
 }
 
 return(-1);
}
